import os
import re

from aws_cdk import core, aws_iam, aws_secretsmanager, aws_s3_deployment, aws_s3, aws_kms
from aws_cdk.aws_lambda import Runtime, Code
from aws_lambda.cloud_formation.lambda_aws_cdk import LambdaFunction
from aws_secret_cdk.vpc_parameters import VPCParameters


class RdsSingleUserPasswordRotation:
    """
    Class which creates a lambda function responsible for RDS single user secret (password) rotation.
    """
    LAMBDA_BACKEND_DEPLOYMENT_PACKAGE = 'SecretsManagerRDSSingleUserRotation.zip'

    def __init__(
            self,
            stack: core.Stack,
            prefix: str,
            secret: aws_secretsmanager.Secret,
            kms_key: aws_kms.IKey,
            vpc_parameters: VPCParameters
    ) -> None:
        """
        Constructor.

        :param stack: A stack in which resources should be created.
        :param prefix: A prefix to give for every resource.
        :param secret: A secret instance which the lambda function should be able to access.
        :param kms_key: Custom or managed KMS key for secret encryption which the
        lambda function should be able to access.
        :param vpc_parameters: VPC parameters for resource (e.g. lambda rotation function) configuration.
        """
        self.rotation_lambda_role = aws_iam.Role(
            scope=stack,
            id=prefix + 'RdsSecretRotationLambdaRole',
            role_name=prefix + 'RdsSecretRotationLambdaRole',
            assumed_by=aws_iam.CompositePrincipal(
                aws_iam.ServicePrincipal("lambda.amazonaws.com"),
                aws_iam.ServicePrincipal("secretsmanager.amazonaws.com"),
        ),
            inline_policies={
                prefix + 'RdsSecretRotationLambdaPolicy': aws_iam.PolicyDocument(
                    statements=[
                        # Secrets are KMS encrypted.
                        # Therefore the lambda function should be able to get this value.
                        aws_iam.PolicyStatement(
                            actions=[
                                'kms:GetSecretValue',
                            ],
                            effect=aws_iam.Effect.ALLOW,
                            resources=[kms_key.key_arn]
                        ),
                        # We enforce lambdas to run in a VPC.
                        # Therefore lambdas need some network interface permissions.
                        aws_iam.PolicyStatement(
                            actions=[
                                'ec2:CreateNetworkInterface',
                                'ec2:ModifyNetworkInterface',
                                'ec2:DeleteNetworkInterface',
                                'ec2:AttachNetworkInterface',
                                'ec2:DetachNetworkInterface',
                                'ec2:DescribeNetworkInterfaces',
                                "logs:CreateLogGroup",
                                "logs:CreateLogStream",
                                "logs:PutLogEvents",
                            ],
                            effect=aws_iam.Effect.ALLOW,
                            resources=['*']
                        ),
                        # Lambda needs to call secrets manager to get secret value in order to update database password.
                        aws_iam.PolicyStatement(
                            actions=[
                                "secretsmanager:DescribeSecret",
                                "secretsmanager:GetSecretValue",
                                "secretsmanager:PutSecretValue",
                                "secretsmanager:UpdateSecretVersionStage"
                            ],
                            effect=aws_iam.Effect.ALLOW,
                            resources=[secret.secret_arn]
                        ),
                        # Not exactly sure about this one.
                        # Despite that, this policy does not impose any security risks.
                        aws_iam.PolicyStatement(
                            actions=[
                                "secretsmanager:GetRandomPassword"
                            ],
                            effect=aws_iam.Effect.ALLOW,
                            resources=['*']
                        )
                    ]
                )
            },
        )

        bucket_name = self.__convert(prefix + 'RdsSecretRotationLambdaDeploymentBucket')
        bucket_deployment = self.__convert(bucket_name + 'Deployment')

        dir_path = os.path.dirname(os.path.realpath(__file__))
        path = os.path.join(dir_path, 'packages')
        deployment_files = aws_s3_deployment.Source.asset(path)

        self.rotation_lambda_deployment_bukcet = aws_s3.Bucket(
            scope=stack,
            id=bucket_name,
            access_control=aws_s3.BucketAccessControl.PRIVATE,
            bucket_name=bucket_name,
            removal_policy=core.RemovalPolicy.DESTROY
        )

        # Deploy lambda function source code.
        self.rotation_lambda_deployment = aws_s3_deployment.BucketDeployment(
            scope=stack,
            id=bucket_deployment,
            destination_bucket=self.rotation_lambda_deployment_bukcet,
            sources=[deployment_files]
        )

        # Create a lambda function responsible for rds password rotation.
        self.rotation_lambda_function = LambdaFunction(
            scope=stack,
            prefix=prefix,
            description='Description',
            memory=128,
            timeout=60,
            handler='lambda_function.lambda_handler',
            runtime=Runtime.PYTHON_2_7,
            role=self.rotation_lambda_role,
            env={
                'SECRETS_MANAGER_ENDPOINT': f'https://secretsmanager.{stack.region}.amazonaws.com'
            },
            security_groups=vpc_parameters.rotation_lambda_security_groups,
            subnets=vpc_parameters.rotation_lambda_subnets,
            vpc=vpc_parameters.rotation_lambda_vpc,
            source_code=Code.from_bucket(
                bucket=self.rotation_lambda_deployment_bukcet,
                key=self.LAMBDA_BACKEND_DEPLOYMENT_PACKAGE,
            )
        ).lambda_function

        # Add some dependencies to ensure the right creation order.
        # Mostly this is not necessary because CloudFormation is smart enough.
        self.rotation_lambda_function.node.add_dependency(self.rotation_lambda_deployment_bukcet)
        self.rotation_lambda_function.node.add_dependency(self.rotation_lambda_deployment)
        self.rotation_lambda_function.node.add_dependency(self.rotation_lambda_role)

    @staticmethod
    def __convert(name: str) -> str:
        """
        Converts CamelCase string to pascal-case where underscores are dashes.
        This is required due to S3 not supporting capital letters or underscores.
        """
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()



import json

from typing import Optional, Union
from aws_cdk import aws_secretsmanager, core, aws_kms, aws_lambda, aws_rds
from aws_cdk.aws_secretsmanager import SecretStringGenerator
from aws_secret_cdk.rds_single_user_password_rotation import RdsSingleUserPasswordRotation
from aws_secret_cdk.secret_key import SecretKey
from aws_secret_cdk.vpc_parameters import VPCParameters


class RdsSecret:
    """
    Class which creates a whole infrastructure around secret management.
    """
    def __init__(
            self,
            stack: core.Stack,
            prefix: str,
            vpc_parameters: VPCParameters,
            database: Union[aws_rds.CfnDBInstance, aws_rds.CfnDBCluster],
            kms_key: Optional[aws_kms.Key] = None
    ) -> None:
        """
        Constructor.

        :param stack: A stack in which resources should be created.
        :param prefix: A prefix to give for every resource.
        :param vpc_parameters: VPC parameters for resource (e.g. lambda rotation function) configuration.
        :param database: A database instance for which this secret should be applied.
        :param kms_key: Custom or managed KMS key for secret encryption.
        """
        # If KMS key is not present, create one.
        self.kms_key = kms_key or SecretKey(stack, prefix).get_kms_key()

        # This template is sent to a lambda function that executes secret rotation.
        # If you choose to change this template, make sure you change lambda
        # function source code too.
        template = {
            'engine': database.engine,
            'host': database.attr_endpoint_address,
            'username': database.master_username,
            'password': database.master_user_password,
            'dbname': None,
            'port': database.port
        }

        # Instances and clusters have different attributes.
        if isinstance(database, aws_rds.CfnDBInstance):
            template['dbname'] = database.db_name
        elif isinstance(database, aws_rds.CfnDBCluster):
            template['dbname'] = database.database_name

        # Create a secret instance.
        self.secret = aws_secretsmanager.Secret(
            scope=stack,
            id=prefix + 'RdsSecret',
            description=f'A secret for {prefix}.',
            encryption_key=self.kms_key,
            generate_secret_string=SecretStringGenerator(
                generate_string_key='password',
                secret_string_template=json.dumps(template)
            ),
            secret_name=prefix + 'RdsSecret'
        )

        # Make sure database is fully deployed and configured before creating a secret for it.
        self.secret.node.add_dependency(database)

        # Create a lambda function for secret rotation.
        self.secret_rotation = RdsSingleUserPasswordRotation(
            stack=stack,
            prefix=prefix,
            secret=self.secret,
            kms_key=kms_key,
            vpc_parameters=vpc_parameters
        )

        # Make sure secrets manager can invoke this lambda function.
        self.sm_invoke_permission = aws_lambda.CfnPermission(
            scope=stack,
            id=prefix + 'SecretsManagerInvokePermission',
            action='lambda:InvokeFunction',
            function_name=self.secret_rotation.rotation_lambda_function.function_name,
            principal="secretsmanager.amazonaws.com",
        )

        # Make sure lambda function is created before making its permissions.
        self.sm_invoke_permission.node.add_dependency(self.secret_rotation.rotation_lambda_function)

        # Apply rotation for the secret instance.
        self.rotation_schedule = aws_secretsmanager.RotationSchedule(
            scope=stack,
            id=prefix + 'RotationSchedule',
            secret=self.secret,
            rotation_lambda=self.secret_rotation.rotation_lambda_function,
            automatically_after=core.Duration.days(30)
        )

        # Make sure invoke permission for secrets manager is created before creating a schedule.
        self.rotation_schedule.node.add_dependency(self.sm_invoke_permission)

        # Instances and clusters have different arns.
        if isinstance(database, aws_rds.CfnDBInstance):
            assert database.db_instance_identifier, 'Instance identifier must be specified.'
            target_arn = f'arn:aws:rds:eu-west-1:{stack.account}:db:{database.db_instance_identifier}'
        elif isinstance(database, aws_rds.CfnDBCluster):
            assert database.db_cluster_identifier, 'Cluster identifier must be specified.'
            target_arn = f'arn:aws:rds:eu-west-1:{stack.account}:cluster:{database.db_cluster_identifier}'
        else:
            raise TypeError('Unsupported DB type.')

        # Instances and clusters should have different attachment types.
        if isinstance(database, aws_rds.CfnDBInstance):
            target_type = 'AWS::RDS::DBInstance'
        elif isinstance(database, aws_rds.CfnDBCluster):
            target_type = 'AWS::RDS::DBCluster'
        else:
            raise TypeError('Unsupported DB type.')

        # Attach the secret instance to the desired database.
        self.target_db_attachment = aws_secretsmanager.CfnSecretTargetAttachment(
            scope=stack,
            id=prefix + 'TargetRdsAttachment',
            secret_id=self.secret.secret_arn,
            target_id=target_arn,
            target_type=target_type
        )



from aws_cdk import core, aws_kms


class SecretKey:
    """
    Class which creates KMS CMK.
    """
    def __init__(
            self,
            scope: core.Construct,
            prefix: str,
    ):
        """
        Constructor.

        :param scope: Usually a stack in which this resource should be added.
        :param prefix: Prefix for KMS name.
        """
        self.__kms_key = aws_kms.Key(
            scope=scope,
            id=prefix + 'SecretEncryptionKey',
            alias=prefix + 'SecretEncryptionKey',
            description=f'Key used to encrypt {prefix} secret.',
            enabled=True,
            enable_key_rotation=True,
            removal_policy=core.RemovalPolicy.DESTROY
        )

    def get_kms_key(self) -> aws_kms.Key:
        """
        Returns a KMS key instance.

        :return: KMS CMK.
        """
        return self.__kms_key



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup(
    name='aws_secret_cdk',
    version='2.0.2',
    license='GNU GENERAL PUBLIC LICENSE Version 3',
    packages=find_packages(exclude=['venv', 'test']),
    description='Package to create a SecretsManager\'s secret with auto rotation.',
    long_description=README + '\n\n' + HISTORY,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        'aws-cdk.core',
        'aws-cdk.aws_iam',
        'aws-cdk.aws_ec2',
        'aws-cdk.aws_lambda',
        'aws-cdk.aws_rds',
        'aws-cdk.aws_secretsmanager',
        'aws-cdk.aws_s3_deployment',
        'aws-lambda>=2.1.2,<3.0.0'
    ],
    author='Laimonas Sutkus',
    author_email='laimonas.sutkus@gmail.com',
    keywords='AWS CDK CloudFormation SecretsManager Infrastructure Cloud DevOps',
    url='https://github.com/laimonassutkus/AwsSecretCdk',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
)
